import errno
import functools
import os

from asyncssh import ProcessError
from asyncssh.misc import PermissionDenied
from asyncssh.sftp import SFTPFailure, SFTPNoSuchFile
from fsspec.asyn import sync_wrapper

_NOT_FOUND = os.strerror(errno.ENOENT)
_FILE_EXISTS = os.strerror(errno.EEXIST)

# A copy of SFTP_BLOCK_SIZE (16KB)
BASE_BLOCK_SIZE = 2 ** 14

# Most of the SFTP implementations support reading a 64kb
# and writing a 256kb chunk at a single request. We'll set
# these values with a 16kb margin.
READ_BLOCK_SIZE = BASE_BLOCK_SIZE * 3
WRITE_BLOCK_SIZE = BASE_BLOCK_SIZE * 15


def wrap_exceptions(func):
    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except PermissionDenied as exc:
            raise PermissionError(exc.reason) from exc
        except SFTPNoSuchFile as exc:
            raise FileNotFoundError(errno.ENOENT, _NOT_FOUND) from exc
        except ProcessError as exc:
            message = exc.stderr.strip()
            if message.endswith(_NOT_FOUND):
                raise FileNotFoundError(errno.ENOENT, _NOT_FOUND) from exc
            raise
        except SFTPFailure as exc:
            message = exc.reason
            if message.endswith("already exists"):
                raise FileExistsError(errno.EEXIST, _FILE_EXISTS) from exc
            raise

    return wrapper


def _mirror_method(method):
    async def _method(self, *args, **kwargs):
        wrapped_meth = getattr(self._file, method)
        return await wrapped_meth(*args, **kwargs)

    _method.__name__ = method
    return sync_wrapper(_method)


def as_progress_handler(callback):
    if callback is None:
        return None

    sent_total = False

    def progress_handler(src_path, dst_path, absolute_progress, total_size):
        nonlocal sent_total
        if not sent_total:
            callback.set_size(total_size)
            sent_total = True

        callback.absolute_update(absolute_progress)

    return progress_handler



import asyncio
from contextlib import suppress

from asyncssh.misc import ChannelOpenError

from sshfs.compat import AsyncExitStack

MAX_TIMEOUT = 60 * 60 * 3


class BaseSFTPChannelPool:
    """BaseSFTPChannelPool is a pool manager for SFTP channels created
    by asyncssh client. The pool might operate in two different modes
    depending on the subclass."""

    def __init__(
        self,
        client,
        *,
        max_channels=None,
        timeout=MAX_TIMEOUT,
        unsafe_terminate=True,
        **kwargs,
    ):
        self.client = client

        # This limit might change during the execution to reflect
        # better to the server's capacity to prevent getting too
        # many errors and wasting time on creating failed channels.
        self.max_channels = max_channels

        # When there are no channels available, this is the maximum amount
        # of time that the SFTPChannelPool will wait to retrieve the
        # channel. If nothing gets released within this parameter, then
        # a TimeoutError will be raised. It can be None.
        self.timeout = timeout

        # When the pool is closing, whether to terminate all open
        # connections or raise an error to indicate there are leaks.
        self.unsafe_terminate = unsafe_terminate
        self._stack = AsyncExitStack()

    async def _maybe_new_channel(self):
        # If there is no hard limit or the limit is not hit yet
        # try to create a new channel
        if (
            self.max_channels is None
            or self.active_channels < self.max_channels
        ):
            try:
                return await self._stack.enter_async_context(
                    self.client.start_sftp_client()
                )
            except ChannelOpenError:
                # If we can't create any more channels, then change
                # the hard limit to reflect that so that we don't hit
                # these errors again.
                self.max_channels = self.active_channels

    async def get(self):
        raise NotImplementedError

    async def _cleanup(self):
        ...

    async def close(self):
        if self.active_channels and not self.unsafe_terminate:
            raise RuntimeError(
                f"{type(self).__name__!r} can't be closed while there are active channels"
            )

        async with asyncio.Lock():
            with suppress(Exception):
                await self._cleanup()

            await self._stack.aclose()



import asyncio

from sshfs.compat import asynccontextmanager
from sshfs.pools.base import BaseSFTPChannelPool


class SFTPHardChannelPool(BaseSFTPChannelPool):
    """A SFTP pool implementation that ensures at any moment in time,
    a single SFTP channel might only be used by a single coroutine. When there
    are no more active channels, the ``.get()`` method will block for a channel
    to get released (if ``timeout`` is specified, it will poll for ``timeout``
    seconds until a ``TimeoutError`` is raised)."""

    def __init__(self, *args, **kwargs):
        self._queue = asyncio.Queue(0)
        self._poll = kwargs.pop("poll", True)
        self.active_channels = 0
        super().__init__(*args, **kwargs)

    @asynccontextmanager
    async def get(self):
        channel = None
        if self._queue.empty():
            channel = await self._maybe_new_channel()

        if channel is None:
            if self._queue.qsize() == 0 and not self.active_channels:
                raise ValueError("Can't create any SFTP connections!")

            if self._poll:
                channel = await asyncio.wait_for(
                    self._queue.get(), timeout=self.timeout
                )
            else:
                channel = self._queue.get_nowait()

        self.active_channels += 1
        try:
            yield channel
        finally:
            self.active_channels -= 1
            self._queue.put_nowait(channel)

    async def _cleanup(self):
        while not self._queue.empty():
            self._queue.get_nowait()



import heapq
from collections import Counter

from sshfs.compat import asynccontextmanager
from sshfs.pools.base import BaseSFTPChannelPool


class SFTPSoftChannelPool(BaseSFTPChannelPool):
    """A SFTP pool implementation that allows usage of same channels
    by multiple coroutines and handles the balanced distribution of multiple
    channels from least used to most used. The ``.get()`` method will not block
    unlike the hard pool and no timeouts will happen on the management side."""

    _THRESHOLD = 4

    # Placeholder to use when there are no channels in
    # the counter.
    _NO_CHANNELS = [[None, _THRESHOLD + 1]]

    def __init__(self, *args, **kwargs):
        self._channels = Counter()
        super().__init__(*args, **kwargs)

    @asynccontextmanager
    async def get(self):
        [(least_used_channel, num_connections)] = (
            heapq.nsmallest(1, self._channels.items(), lambda kv: kv[1])
            or self._NO_CHANNELS
        )

        if least_used_channel is None or num_connections >= self._THRESHOLD:
            channel = await self._maybe_new_channel()
            if channel is not None:
                least_used_channel = channel
                num_connections = 0

        if least_used_channel is None:
            raise ValueError("Can't create any SFTP connections!")

        self._channels[least_used_channel] += 1
        try:
            yield least_used_channel
        finally:
            self._channels[least_used_channel] -= 1

    async def _cleanup(self):
        self._channels.clear()

    @property
    def active_channels(self):
        return len(self._channels)

    @active_channels.setter
    def active_channels(self, value):
        return None



import getpass
from contextlib import suppress
from pathlib import Path

from asyncssh.config import SSHClientConfig

SSH_CONFIG = Path("~", ".ssh", "config").expanduser()


def parse_config(
    *, host, user=(), port=(), local_user=None, config_files=None
):
    if config_files is None:
        config_files = [SSH_CONFIG]

    if local_user is None:
        with suppress(KeyError):
            local_user = getpass.getuser()

    return SSHClientConfig.load(
        reload=False,
        last_config=None,
        config_paths=config_files,
        local_user=local_user,
        user=user,
        host=host,
        port=port,
    )



import io

from asyncssh.sftp import _MAX_SFTP_REQUESTS
from fsspec.asyn import sync

from sshfs.utils import (
    READ_BLOCK_SIZE,
    WRITE_BLOCK_SIZE,
    _mirror_method,
    wrap_exceptions,
)


class SSHFile(io.IOBase):
    def __init__(
        self, fs, path, mode="rb", block_size=None, max_requests=None, **kwargs
    ):
        self.fs = fs
        self.loop = fs.loop

        # TODO: support r+ / w+ / a+
        if mode not in {"rb", "wb", "ab"}:
            raise ValueError(f"Unsupported file mode: {mode}")

        self.path = path
        self.mode = mode
        self.max_requests = max_requests or _MAX_SFTP_REQUESTS

        if block_size is None:
            # "The OpenSSH SFTP server will close the connection
            # if it receives a message larger than 256 KB, and
            # limits read requests to returning no more than
            # 64 KB."
            #
            # We are going to use the maximum block_size possible
            # with a 16KB margin (so instead of sending 256 KB data,
            # we'll send 240 KB + headers for write requests)

            if self.readable():
                block_size = READ_BLOCK_SIZE
            else:
                block_size = WRITE_BLOCK_SIZE

        # The blocksize is often used with constructs like
        # shutil.copyfileobj(src, dst, length=file.blocksize) and since we are
        # using pipelining, we are going to reflect the total size rather than
        # a size of chunk to our limits.
        self.blocksize = block_size * self.max_requests

        self.kwargs = kwargs

        self._file = sync(self.loop, self._open_file)
        self._closed = False

    @wrap_exceptions
    async def _open_file(self):
        # TODO: this needs to keep a reference to the
        # pool as well, otherwise we might broke our
        # guarantee for the hard pool since the file
        # will still be using that channel to perform
        # it's operations but the pool it thinking this
        # channel is freed.
        async with self.fs._pool.get() as channel:
            return await channel.open(
                self.path,
                self.mode,
                block_size=self.blocksize // self.max_requests,
                max_requests=self.max_requests,
            )

    read = _mirror_method("read")
    seek = _mirror_method("seek")
    tell = _mirror_method("tell")

    write = _mirror_method("write")
    fsync = _mirror_method("fsync")
    truncate = _mirror_method("truncate")

    _close = _mirror_method("close")

    def readable(self):
        return "r" in self.mode

    def writable(self):
        return not self.readable()

    def close(self):
        if self._closed:
            return None

        self._close()
        self._closed = True

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()

    @property
    def closed(self):
        return self._closed



import asyncio
import posixpath
import shlex
import stat
import threading
import weakref
from contextlib import suppress
from datetime import datetime

import asyncssh
from asyncssh.sftp import SFTPOpUnsupported
from fsspec.asyn import (
    AsyncFileSystem,
    async_methods,
    fsspec_loop,
    sync,
    sync_wrapper,
)

from sshfs.compat import AsyncExitStack
from sshfs.file import SSHFile
from sshfs.pools import SFTPSoftChannelPool
from sshfs.utils import (
    READ_BLOCK_SIZE,
    WRITE_BLOCK_SIZE,
    as_progress_handler,
    wrap_exceptions,
)

async_methods.append("_mv")

# Always allocate 2 channels for shell operations
# and the rest (generally 8) for SFTP.
_SHELL_CHANNELS = 2
_DEFAULT_MAX_SESSIONS = 10


class SSHFileSystem(AsyncFileSystem):
    def __init__(
        self,
        host,
        *,
        pool_type=SFTPSoftChannelPool,
        **kwargs,
    ):
        """
        Implementation of the SFTP/SSH protocols for the fsspec.

        Parameters
        ----------
        host: str
            SSH host to connect.
        **kwargs: Any
            Any option that will be passed to either the top level `AsyncFileSystem`
            or the `asyncssh.connect`.
        pool_type: sshfs.pools.base.BaseSFTPChannelPool
            Pool manager to use (when doing concurrent operations together,
            pool managers offer the flexibility of prioritizing channels
            and deciding which to use).
        """

        super().__init__(self, **kwargs)

        _client_args = kwargs.copy()
        _client_args.setdefault("known_hosts", None)

        max_sessions = kwargs.get("max_sessions", _DEFAULT_MAX_SESSIONS)
        assert max_sessions > _SHELL_CHANNELS

        with fsspec_loop():
            self._stack = AsyncExitStack()
            # This might not be thread-safe, though fsspec should probably
            # handle those cases on it's loop.
            self._client_lock = asyncio.Semaphore(_SHELL_CHANNELS)

        self.active_executors = 0
        self._client, self._pool = self.connect(
            host,
            pool_type,
            max_sftp_channels=max_sessions - _SHELL_CHANNELS,
            **_client_args,
        )
        weakref.finalize(
            self, sync, self.loop, self._finalize, self._pool, self._stack
        )

    @wrap_exceptions
    async def _connect(
        self, host, pool_type, max_sftp_channels, **client_args
    ):
        _raw_client = asyncssh.connect(host, **client_args)
        client = await self._stack.enter_async_context(_raw_client)
        pool = pool_type(client, max_channels=max_sftp_channels)
        return client, pool

    connect = sync_wrapper(_connect)

    async def _finalize(self, pool, stack):
        await pool.close()

        # If an error occurs while the SSHFile is trying to
        # open the native file, then the client might get broken
        # due to partial initalization. We are just going to ignore
        # the errors that arises on the finalization layer
        with suppress(BrokenPipeError):
            await stack.aclose()

    @property
    def client(self):
        assert self._client is not None
        return self._client

    def _decode_attributes(self, attributes):
        if stat.S_ISDIR(attributes.permissions):
            kind = "directory"
        elif stat.S_ISREG(attributes.permissions):
            kind = "file"
        elif stat.S_ISLNK(attributes.permissions):
            kind = "link"
        else:
            kind = "unknown"

        return {
            "size": attributes.size,
            "type": kind,
            "gid": attributes.gid,
            "uid": attributes.uid,
            "time": datetime.utcfromtimestamp(attributes.atime),
            "mtime": datetime.utcfromtimestamp(attributes.mtime),
            "permissions": attributes.permissions,
        }

    @wrap_exceptions
    async def _info(self, path, **kwargs):
        async with self._pool.get() as channel:
            attributes = await channel.stat(path)

        info = self._decode_attributes(attributes)
        path = path.rstrip("/")
        if info["type"] == "directory":
            path += "/"
        info["name"] = path
        return info

    @wrap_exceptions
    async def _mv(self, lpath, rpath, **kwargs):
        async with self._pool.get() as channel:
            with suppress(SFTPOpUnsupported):
                return await channel.posix_rename(lpath, rpath)

        # Some systems doesn't natively support posix_rename
        # which is an extension to the original SFTP protocol.
        # In that case we are going to copy the file and delete
        # it.

        try:
            await self._cp_file(lpath, rpath)
        finally:
            await self._rm_file(lpath)

    @wrap_exceptions
    async def _put_file(
        self,
        lpath,
        rpath,
        block_size=WRITE_BLOCK_SIZE,
        callback=None,
        **kwargs,
    ):
        await self._makedirs(self._parent(rpath), exist_ok=True)
        async with self._pool.get() as channel:
            await channel.put(
                lpath,
                rpath,
                block_size=block_size,
                progress_handler=as_progress_handler(callback),
            )

    @wrap_exceptions
    async def _get_file(
        self, lpath, rpath, block_size=READ_BLOCK_SIZE, callback=None, **kwargs
    ):
        async with self._pool.get() as channel:
            await channel.get(
                lpath,
                rpath,
                block_size=block_size,
                progress_handler=as_progress_handler(callback),
            )

    @wrap_exceptions
    async def _cp_file(self, lpath, rpath, **kwargs):
        cmd = f"cp {shlex.quote(lpath)} {shlex.quote(rpath)}"
        await self._execute(cmd)

    @wrap_exceptions
    async def _ls(self, path, detail=False, **kwargs):
        async with self._pool.get() as channel:
            file_attrs = await channel.readdir(path)

        infos = []
        for file_attr in file_attrs:
            if file_attr.filename in ["", ".", ".."]:
                continue
            info = self._decode_attributes(file_attr.attrs)
            info["name"] = posixpath.join(path, file_attr.filename)
            infos.append(info)

        # TODO: listings cache
        if detail:
            return infos
        else:
            return [info["name"] for info in infos]

    @wrap_exceptions
    async def _mkdir(
        self, path, *, create_parents=True, permissions=511, **kwargs
    ):
        if create_parents:
            return await self._makedirs(path, exist_ok=True)

        attrs = asyncssh.SFTPAttrs(permissions=permissions)
        async with self._pool.get() as channel:
            await channel.mkdir(path, attrs=attrs)

    @wrap_exceptions
    async def _makedirs(
        self, path, *, exist_ok=False, permissions=511, **kwargs
    ):
        attrs = asyncssh.SFTPAttrs(permissions=permissions)
        async with self._pool.get() as channel:
            await channel.makedirs(path, exist_ok=exist_ok, attrs=attrs)

    mkdir = sync_wrapper(_mkdir)
    makedirs = sync_wrapper(_makedirs)

    @wrap_exceptions
    async def _rm_file(self, path, **kwargs):
        async with self._pool.get() as channel:
            await channel.unlink(path)

    @wrap_exceptions
    async def _rmdir(
        self,
        path,
        recursive=False,
        ignore_errors=False,
        on_error=None,
        **kwargs,
    ):
        async with self._pool.get() as channel:
            if recursive:
                await channel.rmtree(
                    path, ignore_errors=ignore_errors, onerror=on_error
                )
            else:
                await channel.rmdir(path)

    async def _rm(self, path, recursive=False, **kwargs):
        if isinstance(path, str):
            path = [path]

        coros = []
        for sub_path in path:
            if await self._isdir(sub_path):
                coro = self._rmdir(sub_path, recursive, **kwargs)
            else:
                coro = self._rm_file(sub_path)
            coros.append(coro)

        await asyncio.gather(*coros)

    @wrap_exceptions
    async def _checksum(self, path):
        system = await self._get_system()
        if system == "Linux":
            command = "md5sum"
            part = 0
        elif system == "Darwin":
            command = "md5"
            part = -1
        else:
            raise ValueError(f"{system!r} doesn't support checksum operation")

        cmd = f"{command} {shlex.quote(path)}"
        result = await self._execute(cmd)

        parts = result.stdout.strip().split()
        assert len(parts) >= 1

        checksum = parts[part]
        assert len(checksum) == 32
        return checksum

    @wrap_exceptions
    async def _get_system(self):
        result = await self._execute("uname")
        return result.stdout.strip()

    checksum = sync_wrapper(_checksum)
    get_system = sync_wrapper(_get_system)

    async def _execute(self, *args, **kwargs):
        """Execute a shell command on the host system
        and return the result."""
        kwargs.setdefault("check", True)
        async with self._client_lock:
            return await self.client.run(*args, **kwargs)

    execute = sync_wrapper(_execute)

    def _open(self, path, *args, **kwargs):
        return SSHFile(self, path, *args, **kwargs)



__all__ = ["AsyncExitStack", "asynccontextmanager"]

try:
    from contextlib import AsyncExitStack
except ImportError:
    from async_exit_stack import AsyncExitStack

try:
    from contextlib import asynccontextmanager
except ImportError:
    # Unfortunately there isn't any backport of asynccontextmanager,
    # so we have to implement a fairly basic version for our needs until
    # we drop support for 3.6
    # See: https://github.com/python/cpython/blob/fee96422e6f0056561cf74fef2012cc066c9db86/Lib/contextlib.py#L164-L207
    from functools import wraps

    class _AsyncGeneratorContextManager:
        def __init__(self, func, args, kwargs):
            self.gen = func(*args, **kwargs)

        async def __aenter__(self):
            try:
                return await self.gen.__anext__()
            except StopAsyncIteration:
                raise RuntimeError from None

        async def __aexit__(self, typ, value, traceback):
            if typ is None:
                try:
                    await self.gen.__anext__()
                except StopAsyncIteration:
                    return
                else:
                    raise RuntimeError("generator didn't stop")
            else:
                if value is None:
                    value = typ()
                try:
                    await self.gen.athrow(typ, value, traceback)
                    raise RuntimeError("generator didn't stop after athrow()")
                except StopAsyncIteration as exc:
                    return exc is not value
                except RuntimeError as exc:
                    if exc is value:
                        return False
                    if isinstance(value, (StopIteration, StopAsyncIteration)):
                        if exc.__cause__ is value:
                            return False
                    raise
                except BaseException as exc:
                    if exc is not value:
                        raise

    def asynccontextmanager(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return _AsyncGeneratorContextManager(func, args, kwargs)

        return wrapper
